#!/usr/bin/perl
#
# Apply heuristics to geolocate conference and author locations
#
# $Id: geoloc.pl,v 1.56 2010/11/19 12:20:13 panos Exp $
#

$data_dir = 'paper/analysis';

use DBI;
use Getopt::Std;

$dbh = DBI->connect("dbi:SQLite:dbname=$data_dir/names.db");
$dbh->{sqlite_unicode} = 1;

getopts('DgGh:ilprt:') || die "usage: $0 [-DgGilprt] [-h list] [-t test]
-g	geolocate
-h 	specify comma-separated list of heuristics to run
-i	initialize database schema, read geodata and papers
-l	limit data reads and geolocation to small samples
-p	read papers
-r	reset geolocated papers
-t re	run the test specified by the re. The following tests are available: C|S|b|c|u
-G	populate the GDP per capita data
-D	populate country population density data
";

$opt_t = 'xyzzyxyzy' if (!defined($opt_t));

if ($opt_l) {
	$read_sample = 500;
	$heur_limit = 1000;
} else {
	$read_sample = -1;
	$heur_limit = 2000000000;
}

# Live dangerously when loading a lot of data
if ($opt_i || $opt_p) {
	$dbh->do('PRAGMA journal_mode = off;') || die;
	$dbh->do('PRAGMA synchronous = off;') || die;
}

if ($opt_i) {
	populate_gazetteer();
	populate_papers();
	populate_gdp();
	populate_density();
}

populate_gdp() if ($opt_G);

populate_density() if ($opt_D);

read_papers() if ($opt_p);

if ($opt_r) {
	$dbh->do('UPDATE Papers SET confLocId=NULL, corrLocId=NULL, confLocHeur=NULL, corrLocHeur=NULL, CO2=NULL;') || die;
}

print "Testing is enabled\n" if ($opt_l);

if ($opt_t || $opt_g) {
	prepare_statements();
	prepare_hashes();
}

if ($opt_g) {
	open(LOG, '>log') || die;


	conf_name_state_1()	if (!defined($opt_h) || $opt_h =~ m/\b1\b/);
	conf_name_2()		if (!defined($opt_h) || $opt_h =~ m/\b2\b/);
	conf_name_3()		if (!defined($opt_h) || $opt_h =~ m/\b3\b/);
	conf_name_country_6()	if (!defined($opt_h) || $opt_h =~ m/\b6\b/);
	corr_name_country_4()	if (!defined($opt_h) || $opt_h =~ m/\b4\b/);
}

# Write the specified line in a log file and or console
sub logline
{
	my ($line) = @_;
	print LOG $line, "\n";
	print STDERR $line, "\n";
}

sub populate_gazetteer
{
	create_gazetteer_schema();
	read_divisions();

	$name_select = $dbh->prepare('SELECT id FROM Names where Name = ?');
	$name_insert = $dbh->prepare('INSERT INTO Names(name) VALUES(?)');

	read_us_states();
	read_world();
	read_us();
	read_large_us_cities();

	# Create a more sensible default name for each location
	# Probably a perfect solution can't be found with the existing data
	$dbh->do(q{
		UPDATE Locations
		SET nameId=(
		SELECT DISTINCT nameid FROM LocationNames
		WHERE (type='C' OR type='D' OR type='N' OR type='U') AND nodiacritics
		AND LocationNames.locationid = Locations.id);
	}) || die;

}

sub populate_papers
{
	create_paper_schema();
	read_papers();
	read_doi_prefixes();
}

sub populate_gdp
{
	$dbh->do('DROP TABLE GDPPC');
	$dbh->do('CREATE TABLE GDPPC(
		id INTEGER PRIMARY KEY ASC,
		name TEXT UNIQUE ON CONFLICT IGNORE,
		GDPPC REAL);') || die;
	# This uses Perl's default UTF8 encoding
	open(IN, "$data_dir/gdp-per-capita.txt") || die;
	my $i = 0;
	my $oldpc = 100;
	my $sth = $dbh->prepare("INSERT INTO GDPPC(name, GDPPC) VALUES(?, ?);");

	while(<IN>) {
		chop;
		s/(\d),(\d)/$1$2/g;
		@fields = split(/\t/, $_);
		$sth->execute($fields[0], $fields[5]) or die;
		last if ($i++ == $read_sample);
		my $pc = int($i / 185 * 100);
		if ($pc != $oldpc) {
			print STDERR "\rRead GDP: $pc%";
			$oldpc = $pc;
		}
	}
	print STDERR "\n";
}

sub populate_density
{
	$dbh->do('DROP TABLE POPDENSITY');
	$dbh->do('CREATE TABLE POPDENSITY(
		id INTEGER PRIMARY KEY ASC,
		name TEXT UNIQUE ON CONFLICT IGNORE,
		DENSITY REAL);') || die;
	# This uses Perl's default UTF8 encoding
	open(IN, "$data_dir/pop-density.txt") || die;
	my $i = 0;
	my $oldpc = 100;
	my $sth = $dbh->prepare("INSERT INTO POPDENSITY(name, DENSITY) VALUES(?, ?);");

	while(<IN>) {
		chop;
		s/(\d),(\d)/$1$2/g;
		@fields = split(/\t/, $_);
		$sth->execute($fields[0], $fields[2] + 0) or die;
		last if ($i++ == $read_sample);
		my $pc = int($i / 203 * 100);
		if ($pc != $oldpc) {
                  	print STDERR "\rRead density: $pc%";
			$oldpc = $pc;
		}
	}
	# The other databases have 'United States'
	$dbh->do(q{update popdensity set name = 'United States' where name = 'United States of America';});
        # The following insert is not needed any more as it has been entered
        # into the source table.
	# Division of population / total area from:
	# https://www.cia.gov/library/publications/the-world-factbook/geos/tw.html
	# $dbh->do(q{INSERT INTO POPDENSITY(name, DENSITY) VALUES('Taiwan', 640);});
	print STDERR "\n";
}

# Given a name insert it if needed in the Names table and return its code
sub name_code
{
	my ($name) = @_;
	$name_select->execute($name);
	my $id;
	if (defined($id = ($name_select->fetchrow_array)[0])) {
		return $id;
	} else {
		$name_insert->execute($name);
		# Non-standard SQLite function
		return $dbh->last_insert_id(undef, undef, undef, undef);
	}
}


# Format described in
# http://earth-info.nga.mil/gns/html/gis_countryfiles.htm
sub read_world
{
	# This uses Perl's default UTF8 encoding
	open(IN, "$data_dir/populated-world.txt") || die;
	my $i = 0;
	my $oldpc = 100;
	my $loc_select = $dbh->prepare('SELECT id FROM Locations WHERE extId = ?');
	my $loc_insert = $dbh->prepare('INSERT INTO Locations(lat, long, dsg, pc, cc1, adm1, nameId, extId) VALUES(?, ?, ?, ?, ?, ?, ?, ?)');
	my $ln_insert = $dbh->prepare('INSERT INTO LocationNames(locationId, NameId, type, lCode, nodiacritics) VALUES(?, ?, ?, ?, ?)');

	while(<IN>) {
		my ($rc, $ufi, $uni, $lat, $long, $dms_lat, $dms_long, $mgrs,
		$jog, $fc, $dsg, $pc, $cc1, $adm1, $adm2, $pop, $elev, $cc2,
		$nt, $lc, $short, $generic, $sort, $full, $full_nd) =
			split(/\t/, $_);
		$dsg = substr($dsg, 3, 1);
		my $name_id = name_code($full_nd);
		$loc_select->execute($ufi);
		my $loc_id;
		if (!defined($loc_id = ($loc_select->fetchrow_array)[0])) {
			$loc_insert->execute($lat, $long, $dsg, $pc, $cc1, $adm1, $name_id, $ufi);
			$loc_id = $dbh->last_insert_id(undef, undef, undef, undef);
		}

		$ln_insert->execute($loc_id, $name_id, $nt, $lc, 1);
		if ($full ne $full_nd) {
			$name_id = name_code($full);
			$ln_insert->execute($loc_id, $name_id, $nt, $lc, 0);
		}

		last if ($i++ == $read_sample);
		my $pc = int($i / 3704125 * 100);
		if ($pc != $oldpc) {
			print STDERR "\rRead world: $pc%";
			$oldpc = $pc;
		}
	}
	print STDERR "\n";
}

# Format described in
# http://efele.net/maps/fips-10/data/
sub
read_divisions
{
	# This uses Perl's default UTF8 encoding
	open(IN, "$data_dir/fips-all.txt") || die;
	my $i = 0;
	my $oldpc = 100;
	my $sth = $dbh->prepare("INSERT INTO Divisions(country, code, name) VALUES(?, ?, ?)");

	while(<IN>) {
		chop;
		@fields = split(/\_/, $_);
		next if ($fields[2] != 414);		# Current version
		($country, $code) = ($fields[0] =~ m/^(..)(..)$/);
		$sth->execute($country, $code, $fields[7]) or die;
		#print "$country,$code,$fields[7]\n";
		last if ($i++ == $read_sample);
		my $pc = int($i / 6097 * 100);
		if ($pc != $oldpc) {
			print STDERR "\rRead divisions: $pc%";
			$oldpc = $pc;
		}
	}
	print STDERR "\n";
}

sub
read_us_states
{
	open(IN, "$data_dir/us-states.txt") || die;
	my $sth = $dbh->prepare("INSERT INTO Divisions(country, code, name) VALUES('US', ?, ?)");

	print STDERR "Read US states\n";
	while(<IN>) {
		chop;
		@fields = split(/\t/, $_);
		$sth->execute($fields[1], $fields[0]) or die;
	}
}


# Remove diacritics according to rules for region 1 listed in
# http://earth-info.nga.mil/gns/html/Char_sets_20060703.pdf
sub strip_diacr
{
	my ($s) = @_;
	# Each page from the document is listed in a separate block

	$s =~ s/\x{10C}/C/g;
	$s =~ s/\x{152}/OE/g;
	$s =~ s/\x{14A}/N/g;
	$s =~ s/\x{150}/O/g;

	$s =~ s/\x{160}/S/g;
	$s =~ s/\x{201d}/"/g;
	$s =~ s/\x{166}/T/g;
	$s =~ s/\x{17D}/Z/g;
	$s =~ s/\x{10D}/c/g;
	$s =~ s/\x{153}/oe/g;

	$s =~ s/\x{14B}/n/g;
	$s =~ s/\x{151}/o/g;
	#$s =~ s/\x{b0}//g; (NOP?)
	$s =~ s/\x{161}/s/g;
	$s =~ s/\x{b7}/./g;
	$s =~ s/\x{167}/t/g;
	$s =~ s/\x{17E}/z/g;
	$s =~ s/\x{C0}/A/g;

	$s =~ s/\x{C1}/A/g;
	$s =~ s/\x{C2}/A/g;
	$s =~ s/\x{C3}/A/g;
	$s =~ s/\x{C4}/A/g;
	$s =~ s/\x{C5}/A/g;
	$s =~ s/\x{C6}/AE/g;
	$s =~ s/\x{C7}/C/g;
	$s =~ s/\x{C8}/E/g;
	$s =~ s/\x{C9}/E/g;
	$s =~ s/\x{CA}/E/g;
	$s =~ s/\x{CB}/E/g;
	$s =~ s/\x{CC}/I/g;
	$s =~ s/\x{CD}/I/g;
	$s =~ s/\x{CE}/I/g;
	$s =~ s/\x{CF}/I/g;
	$s =~ s/\x{D0}/Dh/g;
	$s =~ s/\x{D1}/N/g;
	$s =~ s/\x{D2}/O/g;
	$s =~ s/\x{D3}/O/g;
	$s =~ s/\x{D4}/O/g;
	$s =~ s/\x{D5}/O/g;
	$s =~ s/\x{D6}/O/g;
	$s =~ s/\x{170}/U/g;

	$s =~ s/\x{D8}/O/g;
	$s =~ s/\x{D9}/U/g;
	$s =~ s/\x{DA}/U/g;
	$s =~ s/\x{DB}/U/g;
	$s =~ s/\x{DC}/U/g;
	$s =~ s/\x{DD}/Y/g;
	$s =~ s/\x{DE}/Th/g;
	$s =~ s/\x{DF}/ss/g;
	$s =~ s/\x{E0}/a/g;
	$s =~ s/\x{E1}/a/g;
	$s =~ s/\x{E2}/a/g;
	$s =~ s/\x{E3}/a/g;
	$s =~ s/\x{E4}/a/g;
	$s =~ s/\x{E5}/a/g;
	$s =~ s/\x{E6}/ae/g;
	$s =~ s/\x{E7}/c/g;
	$s =~ s/\x{E8}/e/g;
	$s =~ s/\x{E9}/e/g;
	$s =~ s/\x{EA}/e/g;
	$s =~ s/\x{EB}/e/g;
	$s =~ s/\x{EC}/i/g;
	$s =~ s/\x{ED}/i/g;
	$s =~ s/\x{EE}/i/g;

	$s =~ s/\x{EF}/i/g;
	$s =~ s/\x{111}/dh/g;
	$s =~ s/\x{F1}/n/g;
	$s =~ s/\x{F2}/o/g;
	$s =~ s/\x{F3}/o/g;
	$s =~ s/\x{F4}/o/g;
	$s =~ s/\x{F5}/o/g;
	$s =~ s/\x{F6}/o/g;
	$s =~ s/\x{171}/u/g;
	$s =~ s/\x{F8}/o/g;
	$s =~ s/\x{F9}/u/g;
	$s =~ s/\x{FA}/u/g;
	$s =~ s/\x{FB}/u/g;
	$s =~ s/\x{FC}/u/g;
	$s =~ s/\x{FD}/y/g;
	$s =~ s/\x{FE}/th/g;
	$s =~ s/\x{F0}/dh/g;
	return $s;
}

# Format is
# 0:FEATURE_ID|1:FEATURE_NAME|2:FEATURE_CLASS|3:STATE_ALPHA|4:STATE_NUMERIC|5:COUNTY_NAME|6:COUNTY_NUMERIC|7:PRIMARY_LAT_DMS|8:PRIM_LONG_DMS|9:PRIM_LAT_DEC|10:PRIM_LONG_DEC|11:SOURCE_LAT_DMS|12:SOURCE_LONG_DMS|13:SOURCE_LAT_DEC|14:SOURCE_LONG_DEC|15:ELEVATION|16:MAP_NAME|17:DATE_CREATED|18:DATE_EDITED
sub
read_us
{
	open(IN, "<:encoding(ucs-2le)", "$data_dir/populated-us.txt") || die;
	my $i = 0;
	my $oldpc = 100;
	my $loc_insert = $dbh->prepare(q{INSERT INTO Locations(lat, long, cc1, adm1, nameId, extId) VALUES(?, ?, 'US', ?, ?, ?)});
	my $ln_insert = $dbh->prepare('INSERT INTO LocationNames(locationId, NameId, type, lCode, nodiacritics) VALUES(?, ?, ?, ?, ?)');

	my $dummy = <IN>;
	my $washington_id;
	while(<IN>) {
		@fields = split(/\|/, $_);

		my $full = $fields[1];
		my $full_nd = strip_diacr($fields[1]);
		my $name_id = name_code($full_nd);
		$loc_insert->execute($fields[9], $fields[10], $fields[3], $name_id, $fields[0]) or die;
		my $loc_id = $dbh->last_insert_id(undef, undef, undef, undef);
		$washington_id= $loc_id if ($full eq 'Washington' && $fields[3] eq 'DC');

		$ln_insert->execute($loc_id, $name_id, 'U', 'eng', 1);
		if ($full ne $full_nd) {
			$name_id = name_code($full);
			$ln_insert->execute($loc_id, $name_id, 'U', 'eng', 0);
		}
		last if ($i++ == $read_sample);
		my $pc = int($i / 195619 * 100);
		if ($pc != $oldpc) {
			print STDERR "\rRead US: $pc%";
			$oldpc = $pc;
		}
	}
	$dbh->do(q{UPDATE Locations SET dsg='C' WHERE id=?}, undef, $washington_id);
	print STDERR "\n";
}

# Format is
# rank,name city,state,...
sub
read_large_us_cities
{
	open(IN, "$data_dir/large-us.csv") || die;
	my $i = 0;
	my $oldpc = 100;
	my $loc_find = $dbh->prepare(q{SELECT locationId as id FROM World
		WHERE country='US' and lname=? and dname=?});
	my $loc_update = $dbh->prepare(q{UPDATE Locations SET pc=1 WHERE id=?});

	for (my $i = 0; $i < 4; $i++) {
		my $dummy = <IN>;
	}
	$i = 0;
	while(<IN>) {
		last if (/^\"Note/);
		my($rank, $name, $state) = split(/\,/, $_);

		$name =~ s/ (County|metropolitan|city).*\(balance\)//;
		$name =~ s/ (city|municipality|town|urban county|CDP)//;
		$name =~ s/San Buenaventura \(Ventura\)/San Buenaventura-Ventura/;
		@names = split(/[\/\-]/, $name);
		for $name (@names) {
			$name =~ s/St\./Saint/;
			$name =~ s/Boise City/Boise/;
			$loc_find->execute($name, $state);
			my $r = fetchrow_count($loc_find);
			if ($r->{'n'} == 0) {
				# logline("No name state: [$name] [$state]");
				next;
			} elsif ($r->{'n'} > 1) {
				# logline("No unique name state: [$name] [$state]");
				next;
			}

			$loc_update->execute($r->{'id'}) || die;
		}
		last if ($i++ == $read_sample);
		my $pc = int($i / 280 * 100);
		if ($pc != $oldpc) {
			print STDERR "\rRead large US cities: $pc%";
			$oldpc = $pc;
		}
	}
	print STDERR "\n";
}

sub
create_gazetteer_schema
{
	$dbh->do('CREATE TABLE Divisions(
		country CHAR(2),
		code CHAR(2),
		name TEXT,
		PRIMARY KEY(country, code));') || die;

	$dbh->do('CREATE INDEX dName on Divisions(name);') || die;
	$dbh->do('CREATE INDEX dCountry on Divisions(country);') || die;
	$dbh->do('CREATE INDEX dCode on Divisions(code);') || die;

	$dbh->do('CREATE TABLE Locations(
		id INTEGER PRIMARY KEY ASC,	-- Internal identifier
		lat FLOAT,			-- Lattitude
		long FLOAT,			-- Longitude
		dsg CHAR(1),			-- Designation (fourth PPL letter from NGA DSG - C: Capital)
		pc INTEGER(1),			-- Populated place classification (NGA 1-5)
		cc1 CHAR(2)			-- Primary country code
			REFERENCES Divisions(country),
		adm1 CHAR(2)			-- First-order administrative code (NGA or US state)
			REFERENCES Divisions(code),
		nameId INTEGER			-- Its primary display name
			REFERENCES Names(id),
		extId INTEGER			-- NGA UFI or USGS id (for cc1 = US)
		);') || die;

	$dbh->do('CREATE INDEX lDsg on Locations(dsg);') || die;
	$dbh->do('CREATE INDEX lPc on Locations(pc);') || die;
	$dbh->do('CREATE INDEX lCc1 on Locations(cc1);') || die;
	$dbh->do('CREATE INDEX lAdm1 on Locations(adm1);') || die;
	$dbh->do('CREATE INDEX lExtid on Locations(extId);') || die;

	$dbh->do('CREATE TABLE Names(
		id INTEGER PRIMARY KEY ASC,	-- Internal identifier
		name TEXT			-- The corresponding name
		);') || die;

	$dbh->do('CREATE INDEX nName on Names(name);') || die;

	$dbh->do('CREATE TABLE LocationNames(
		id INTEGER PRIMARY KEY ASC,	-- Internal identifier
		locationId INTEGER		-- Location
			REFERENCES Locations(id),
		nameId INTEGER			-- Its name
			REFERENCES Names(id),
		type CHAR(2),			-- NGA NT field for this relation; U for USGS names
		lCode CHAR(3),			-- ISO639-3 language code (e.g. eng) (NGA LC field)
		nodiacritics BOOLEAN		-- True if in this relation name is sans diacratics (NGA FULL_NAME_ND or USGS stripped by us)
		);') || die;
	$dbh->do('CREATE INDEX lnNameId on LocationNames(nameId);') || die;
	$dbh->do('CREATE INDEX lnLocationId on LocationNames(locationId);') || die;

	$dbh->do('
		CREATE VIEW World AS SELECT *,
		names.name as lName, divisions.name as dName
		FROM locationnames
		INNER JOIN locations ON locationnames.locationid = locations.id
		INNER JOIN names ON names.id = locationnames.nameid
		INNER JOIN Divisions on Locations.cc1 = Divisions.country AND
		Locations.adm1 = Divisions.code;
	') || die;
}

sub
create_paper_schema
{
	$dbh->do('CREATE TABLE Conferences(
		id INTEGER PRIMARY KEY ASC,
		name TEXT UNIQUE ON CONFLICT IGNORE);') || die;
	$dbh->do('CREATE INDEX cName on Conferences(name);') || die;

	$dbh->do('CREATE TABLE Sponsors(
		id INTEGER PRIMARY KEY ASC,
		name TEXT UNIQUE ON CONFLICT IGNORE);') || die;
	$dbh->do('CREATE INDEX sName on Sponsors(name);') || die;

	$dbh->do('CREATE TABLE PaperSponsors(
		paperId INTEGER REFERENCES Papers(id),
		sponsorId INTEGER REFERENCES Sponsors(id),
		PRIMARY KEY(paperId, sponsorId));') || die;

	$dbh->do('CREATE TABLE Papers(
		id INTEGER PRIMARY KEY ASC,			-- Internal identifier
		confLoc TEXT,					-- Conference location as specified in Scopus
		confLocId INTEGER REFERENCES Locations(Id),	-- Geolocated conference location
		confLocHeur INTEGER(1),				-- Heuristic used for obtaining geolocated conference location
		corrAddr TEXT,					-- Correspondence address as specified in Scopus
		corrLocId INTEGER REFERENCES Locations(Id),	-- Geolocated correspondence address
		corrLocHeur INTEGER(1),				-- Heuristic used for obtaining geolocated correspondence address
		conference INTEGER REFERENCES Conferences(id),	-- Pointer to conference
		citations INTEGER(2),				-- Citations to this paper
		bDate DATE,					-- Conference begin date
		eDate DATE,					-- Conference end date
		pYear INTEGER(2),				-- Publication year
		nPages INTEGER(2),				-- Number of pages
		doi TEXT,					-- Digital object identifier
		eid TEXT,					-- Scopus identifier
		CO2 REAL					-- Calculated CO2 emissions for this paper
		);') || die;

	$dbh->do('CREATE INDEX pConfLocId on Papers(confLocId);') || die;
	$dbh->do('CREATE INDEX pCorrLocId on Papers(corrLocId);') || die;
	$dbh->do('CREATE INDEX pEid on Papers(eid);') || die;
	$dbh->do('CREATE INDEX pYear on Papers(pYear);') || die;

	$dbh->do('CREATE TABLE Doi(
		prefix INTEGER(2) PRIMARY KEY,
		publisher TEXT);') || die;
}

sub read_doi_prefixes
{
	# This uses Perl's default UTF8 encoding
	open(IN, "$data_dir/doi-prefix.txt") || die;
	my $i = 0;
	my $oldpc = 100;
	my $sth = $dbh->prepare("INSERT INTO Doi(prefix, publisher) VALUES(?, ?);");

	while(<IN>) {
		chop;
		@fields = split(/\t/, $_);
		$sth->execute($fields[1], $fields[0]) or die;
		last if ($i++ == $read_sample);
		my $pc = int($i / 1452 * 100);
		if ($pc != $oldpc) {
			print STDERR "\rRead doi prefixes: $pc%";
			$oldpc = $pc;
		}
	}
	print STDERR "\n";
}


sub
read_papers
{

	my @files = <$data_dir/ris/*.ris>;
	my $paper_sel = $dbh->prepare('SELECT id FROM Papers WHERE eid = ?;');
	my $paper_ins = $dbh->prepare('INSERT INTO Papers(confLoc, corrAddr, conference, citations, bDate, eDate, pYear, nPages, doi, eid)
		VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?);');
	my $sponsor_ins = $dbh->prepare('INSERT INTO Sponsors(name) VALUES (?);');
	my $sponsor_sel = $dbh->prepare('SELECT id FROM Sponsors WHERE name = ?;');
	my $ps_ins = $dbh->prepare('INSERT INTO PaperSponsors(paperId, sponsorId) VALUES (?, ?);');
	my $conf_ins = $dbh->prepare('INSERT INTO Conferences(name) VALUES (?);');
	my $conf_sel = $dbh->prepare('SELECT id FROM Conferences WHERE name = ?;');
	my %month = (
		'January' =>	1,
		'February' =>	2,
		'March' =>	3,
		'April' =>	4,
		'May' =>	5,
		'June' =>	6,
		'July' =>	7,
		'August' =>	8,
		'September' =>	9,
		'October' =>	10,
		'November' =>	11,
		'December' =>	12
	);


	for $f (@files) {
		# RIS files use Perl's default UTF8 encoding
		open(IN, $f) || die;
		while (<IN>) {
			chop;
			if (s/^N1  \- Correspondence Address: //) {
				$corr_addr = $_;
			} elsif (s/^CY  \- //) {
				$conf_loc = $_;
			# JA  - Int. Conf. Wirel. Commun., Netw. Mob. Comput., WiCOM
			# Abreviated conference title (no year)
			} elsif (s/^JA  \- //) {
				$conference = $_;
			# N1  - Sponsors: IEEE CAS/CANDE; IEEE Circuits and Systems Society; CANDE; ACM
			} elsif (s/^N1  \- Sponsors://) {
				s/\s*\;\s*/;/g;
				s/^\s+//;
				s/\s+$//;
				@sponsors = split(/\;/, $_);
			# N1  - doi: 10.1109/ICCAD.2008.4681654
			} elsif (s/^N1  \- doi: //) {
				$doi = $_;
			} elsif (m/^N1  \- Cited By .*\:\s*(\d+)/) {
				$citations = $1;
			# Y2  - 12 October 2008 through 14 October 2008
			} elsif (m/^Y2  \- (\d+)\s+(\w+)\s+(\d+)\s+through\s+(\d+)\s+(\w+)\s+(\d+)/) {
				$bDate = sprintf("%04d-%02d-%02d", $3, $month{$2}, $1);
				$eDate = sprintf("%04d-%02d-%02d", $6, $month{$5}, $4);
			# Publication year
			} elsif (m/PY  \- (\d+)/) {
				$pYear = $1;
			} elsif (m/SP  \- (\d+)/) {
				$sPage = $1;
			} elsif (m/EP  \- (\d+)/) {
				$ePage = $1;
			} elsif (m,^UR  \- http://www.scopus.com/inward/record.url\?eid=([^&]+)\&,) {
				$eid = $1;
			} elsif (/^ER /) {
				# See if paper already in the database
				$paper_sel->execute($eid);
				if (!defined(($paper_sel->fetchrow_array)[0])) {

					# Calculate # of pages
					$nPages = $ePage - $sPage + 1 if defined($sPage) && defined($ePage);

					# Replace conference with its identifier
					if (defined($conference)) {
						$conf_ins->execute($conference) || die;
						$conf_sel->execute($conference) || die;
						$conference = ($conf_sel->fetchrow_array)[0];
					}

					$paper_ins->execute($conf_loc, $corr_addr, $conference, $citations, $bDate, $eDate, $pYear, $nPages, $doi, $eid) || die;
					$paper_sel->execute($eid) || die;

					# Insert the sponsor relations
					my $paper_id = ($paper_sel->fetchrow_array)[0];
					my %seen;
					for my $sponsor (@sponsors) {
						next if ($seen{$sponsor});
						$seen{$sponsor} = 1;
						$sponsor_ins->execute($sponsor) || die;
						$sponsor_sel->execute($sponsor) || die;
						my $sponsor_id = ($sponsor_sel->fetchrow_array)[0];
						$ps_ins->execute($paper_id, $sponsor_id) || die;
					}
					undef %seen;

					print STDERR '.';
					undef $corr_addr;
					undef $conf_loc;
					undef $conference;
					undef @sponsors;
					undef $doi;
					$citations = 0;
					undef $bDate;
					undef $eDate;
					undef $pYear;
					undef $sPage;
					undef $ePage;
					undef $eid;

					undef $nPages;
					return if ($i++ == $read_sample);
				}
			}
		}
		close(IN);
	}
}

# Prepare associative arrays used for geolocation
sub
prepare_hashes
{
	# Two letter codes used like US state names
	# Example: Vancouver, BC
	$state_like = {
		'AB' => {
			'cc1' => 'CA',
			'adm1' => '01'
		},
		'BC' => {
			'cc1' => 'CA',
			'adm1' => '02'
		},
		'NB' => {
			'cc1' => 'CA',
			'adm1' => '04'
		},
		'NS' => {
			'cc1' => 'CA',
			'adm1' => '07'
		},
		'ON' => {
			'cc1' => 'CA',
			'adm1' => '08'
		},
		'QC' => {
			'cc1' => 'CA',
			'adm1' => '10'
		},
		'SA' => {
			'cc1' => 'AS',
			'adm1' => '05'
		},
		'SK' => {
			'cc1' => 'CA',
			'adm1' => '10'
		},

		'CH' => { 'cc1' => 'SZ', },	# Switzerland
		'DK' => { 'cc1' => 'DA', },	# Denmark
		'SE' => { 'cc1' => 'SW', },	# Sweden
		'UK' => { 'cc1' => 'UK', },	# United Kingdom

		# Clashes with Washington
		#'WA' => {
		#	'cc1' => 'AS',
		#	'adm1' => '08'
		# },
	};

	# Country aliases
	$country_alias = {
		'SOUTH KOREA' => 'KOREA, REPUBLIC OF',
		'RUSSIAN FEDERATION' => 'RUSSIA',
		'QUE.' => 'CANADA',
		'QUE' => 'CANADA',
		'CAN' => 'CANADA',
		'ONT.' => 'CANADA',
		'JPN' => 'JAPAN',
		'SCOTLAND' => 'UNITED KINGDOM',
		'WALES' => 'UNITED KINGDOM',
		'ENGLAND' => 'UNITED KINGDOM',
		'AUST' => 'AUSTRALIA',
		'NSW' => 'AUSTRALIA',
		'GER' => 'GERMANY',
		'NETH' => 'NETHERLANDS',
	};

	# See http://en.wikipedia.org/wiki/Canadian_subnational_postal_abbreviations
	# The beginning \b is added by the array's user
	# The trailing must be put only in letter-ending forms
	@canada_subs = (
		{ 'name' => 'AB\b',	'adm1' => '01' },
		{ 'name' => 'AL\b',	'adm1' => '01' },	# Unofficial
		{ 'name' => 'Alta\.',	'adm1' => '01' },	# Deprecated
		{ 'name' => 'Alb\.',	'adm1' => '01' },	# Unofficial
		{ 'name' => 'BC\b',	'adm1' => '02' },
		{ 'name' => 'B\.C\.',	'adm1' => '02' },
		{ 'name' => 'MB\b',	'adm1' => '03' },
		{ 'name' => 'Man\.',	'adm1' => '03' },
		{ 'name' => 'NB\b',	'adm1' => '04' },
		{ 'name' => 'N\.B\.',	'adm1' => '04' },
		{ 'name' => 'NL\b',	'adm1' => '05' },
		{ 'name' => 'Nfld\.',	'adm1' => '05' },
		{ 'name' => 'NF\b',	'adm1' => '05' },
		{ 'name' => 'NS\b',	'adm1' => '07' },
		{ 'name' => 'N\.S\.',	'adm1' => '07' },
		{ 'name' => 'NT\b',	'adm1' => '13' },
		{ 'name' => 'NU\b',	'adm1' => '14' },
		{ 'name' => 'ON\b',	'adm1' => '08' },
		{ 'name' => 'Ont\.',	'adm1' => '08' },
		{ 'name' => 'PE\b',	'adm1' => '09' },
		{ 'name' => 'QC\b',	'adm1' => '10' },
		{ 'name' => 'Que\.',	'adm1' => '10' },
		{ 'name' => 'Qu\x{e9}\.','adm1' => '10' },
		{ 'name' => 'P\.Q\.',	'adm1' => '10' },
		{ 'name' => 'QB\b',	'adm1' => '10' },
		{ 'name' => 'SK\b',	'adm1' => '11' },
		{ 'name' => 'Sask\.',	'adm1' => '11' },
		{ 'name' => 'YT\b',	'adm1' => '12' },
		{ 'name' => 'Yuk\.',	'adm1' => '12' },
	);
}

# Prepare SQL statements used for geolocation
sub
prepare_statements
{
	$q_conference_todo = $dbh->prepare(q{
	  SELECT id, confLoc FROM Papers WHERE confLocId ISNULL and confLoc <> '' and corrAddr <> ''});

	$q_author_todo = $dbh->prepare(q{
	  SELECT id, corrAddr FROM Papers WHERE corrLocId ISNULL and confLoc <> '' and corrAddr <> ''});

	$q_name = $dbh->prepare(q{
		SELECT locationId AS id FROM LocationNames
		INNER JOIN Names ON Names.id = LocationNames.nameId
		WHERE name=?
	  });

	$q_name_country = $dbh->prepare(q{
		SELECT locationId AS id FROM LocationNames
		INNER JOIN Names ON Names.id = LocationNames.nameId
		INNER JOIN Locations ON LocationNames.locationId = Locations.id
		WHERE name=? AND cc1 =?
		});

	$q_name_state_country = $dbh->prepare(q{
		SELECT locationId AS id FROM LocationNames
		INNER JOIN Names ON Names.id = LocationNames.nameId
		INNER JOIN Locations ON LocationNames.locationId = Locations.id
		WHERE name=? AND adm1 = ? AND cc1 = ?
		});

	$q_important_name = $dbh->prepare(q{
		SELECT locationId AS id FROM LocationNames
		INNER JOIN Names ON Names.id = LocationNames.nameId
		INNER JOIN Locations ON LocationNames.locationId = Locations.id
		WHERE name=? AND (pc=1 or dsg='C' or dsg='A')
		});

	$q_important_name_in_country = $dbh->prepare(q{
		SELECT locationId AS id FROM LocationNames
		INNER JOIN Names ON Names.id = LocationNames.nameId
		INNER JOIN Locations ON LocationNames.locationId = Locations.id
		WHERE name=? AND cc1=? AND (pc=1 or dsg='C' or dsg='A')
		});

	$q_country_code = $dbh->prepare(q{
	  SELECT country FROM
	  Divisions
	  WHERE
	  name = ? AND code = '00'});
}

# Return the code for the specified country
sub
country_code
{
	my ($country) = @_;
	$country =~ s/^\s+//;
	$country =~ s/\s+$//;
	$country = uc($country);
	my $proper = $country_alias->{$country};
	$country = $proper if (defined($proper));
	return $country_cache{$country} if (defined($country_cache{$country}));
	return undef if (defined($negative_country_cache{$country}));
	$q_country_code->execute($country);
	my $r = ($q_country_code->fetchrow_array)[0];
	if (defined($r)) {
		$country_cache{$country} = $r;
	} else {
		$negative_country_cache{$country} = 1;
		logline("No country: [$country]");
	}
	return $r;
}

if ('C' =~ m/$opt_t/) {
	print "Country aliases\n";
	print country_code('SCOTLAND'), "\n";
	print country_code('NSW'), "\n";
	print country_code('RUSSIA'), "\n";
	print country_code('Iran'), "\n";
}

# Return an associative array reference containing in n an indication of the number
# of values that can be fetched through the passed handle.
# If n is exactly one, then id contains the returned value's identifier.
# If more than one value is available set n to 999
sub
fetchrow_count
{
	my($handle) = @_;
	my $r = $handle->fetchrow_hashref;
	if (defined($r)) {
		# Must fetch at most all other rows to see if there are other locationids
		# Under SQLite embedding this in the query as
		# COUNT(DISTINCT LocationId) or ORDER BY LocationId
		# delays the queries by fetching all results whereas we want
		# to quit after the first different LocationId
		# The 'b' test runs in 3.39s with ORDER BY and in 1.27s
		# with this implementation
		my $id = $r->{'id'};
		while ($r = $handle->fetchrow_hashref) {
			if (defined($r) && $r->{'id'} != $id) {
				return { 'n' => 999 };
			}
		}
		return {
			'id' => $id,
			'n' => 1
		};
	} else {
		return {
			'n' => 0
		};
	}
}


# Return the location id for the specified name and state
# only if this is unique
# Otherwise return undef
# Update the counters nonunique, unique, notfound
sub
unique_name_state_country
{
	my ($name, $state, $country) = @_;
	if (!defined($country) || ($country ne 'US' && $country ne 'CA')) {
		my $sl = %$state_like->{$state};
		if (defined($sl)) {
			$country = $sl->{'cc1'};
			$state = $sl->{'adm1'};
		} else {
			$country = 'US';
		}
	}
	$q_name_state_country->execute($name, $state, $country);
	print STDERR "Look for [$name] [$state] [$country]\n";
	my $r = fetchrow_count($q_name_state_country);
	if ($r->{'n'} == 0) {
		logline("No name state country: [$name] [$state] [$country]");
		$notfound++;
		return undef;
	} elsif ($r->{'n'} == 1) {
		$unique++;
		return $r->{'id'};
	} else {
		$nonunique++;
		logline("No unique name state country: [$name] [$state] [$country]");
		return undef;
	}
}

if ('S' =~ m/$opt_t/) {
	print "State aliases\n";
	print unique_name_state_country('San Diego', 'CA'), "\n";
	print unique_name_state_country('San Diego', 'CA', 'US'), "\n";
	print unique_name_state_country('Vancouver', 'BC'), "\n";
}

# Return the location id for the specified name and country
# only if this is unique
# Otherwise return undef
# Update the counters nonunique, unique, notfound, notfoundimp, nonuniqueimp, uniqueimp
sub
name_country
{
	my ($name, $country) = @_;
	$q_name_country->execute($name, $country);
	print STDERR "Look for [$name] [$country]\n";
	my $r = fetchrow_count($q_name_country);
	if ($r->{'n'} == 0) {
		logline("No name country: [$name] [$country]");
		$notfound++;
		return undef;
	} elsif ($r->{'n'} == 1) {
		$unique++;
		return $r->{'id'};
	}
	$nonunique++;

	# Try for an important city (capital or PPLC)
	$q_important_name_in_country->execute($name, $country);
	my $r = fetchrow_count($q_important_name_in_country);
	if ($r->{'n'} == 0) {
		logline("No important name in country: [$name] [$country]");
		$notfoundimp++;
		return undef;
	} elsif ($r->{'n'} > 1) {
		logline("No unique important name in country: [$name] [$country]");
		$nonuniqueimp++;
		return undef;
	}
	$foundimp++;
	return $r->{'id'};
}

if ('b' =~ m/$opt_t/) {
	print "Unique name country\n";
	print "Find single ", name_country('Athens', 'GR'), "\n";
	print "Find with PC=1 ", name_country('Athens', 'US'), "\n";
	print "No find ", name_country('Athens', 'ZA'), "\n";
	print "No find unique ", name_country('Springfield', 'US'), "\n";
	print "Find with DSG=A ", name_country('Stockholm', 'SW'), "\n";
	print "Find among two names for same location ", name_country('Montreal', 'CA'), "\n";
}


# Return the location id for the specified name
# only if this is unique
# Otherwise return undef
# Update the counters nonunique, unique, notfound
sub
unique_name
{
	my ($name) = @_;
	$q_name->execute($name);
	# print STDERR "Look for [$name]\n";
	my $r = fetchrow_count($q_name);
	if ($r->{'n'} == 0) {
		logline("No name: [$name]");
		$notfound++;
		return undef;
	} elsif ($r->{'n'} == 1) {
		$unique++;
		return $r->{'id'};
	} else {
		logline("No unique name: [$name]");
		$nonunique++;
		return undef;
	}
}

# Heuristic 1
# Conference location specified as unique name and state
# Example: Anaheim, CA
sub
conf_name_state_1
{
	$q_conference_todo->execute || die;
	my $u = $dbh->prepare(q{
	  UPDATE Papers SET confLocId = ?, confLocHeur = 1 WHERE id = ?});
	$total = $match = $nonunique = $unique = $notfound = 0;
	my @updates;
	while (my $r = $q_conference_todo->fetchrow_hashref) {
		$total++;
		next unless ($r->{confLoc} =~ m/^\s*(.*)\s*\,\s*([A-Z]{2})(\.|\, USA)?$/);
		$match++;
		$name = $1;
		$state = $2;
		$name =~ s/^([A-Z])(\w+)$/$1 . lc($2)/e;
		next unless ($locId = unique_name_state_country($name, $state, undef));
		push(@updates, {
			'locId' => $locId,
			'id' => $r->{id}});
		print STDERR "Located $r->{confLoc}\n";
		if ($total++ >= $heur_limit) {
			$q_conference_todo->finish;
			last;
		}
	}
	print STDERR "Heuristic 1:\ntotal=$total match=$match nonunique=$nonunique unique=$unique notfound=$notfound\n";
	$dbh->do('BEGIN TRANSACTION');
	for my $urow (@updates) {
		$u->execute($urow->{'locId'}, $urow->{'id'}) or die;
	}
	$dbh->do('END TRANSACTION');
}

# Heuristic 2
# Conference location specified as a unique name
# Example: Kuala Lumpur
sub
conf_name_2
{
	$q_conference_todo->execute || die;
	my $u = $dbh->prepare(q{
	  UPDATE Papers SET confLocId = ?, confLocHeur = 2 WHERE id = ?});
	$match = $total = $nonunique = $unique = $notfound = 0;
	my @updates;
	while (my $r = $q_conference_todo->fetchrow_hashref) {
		$total++;
		next if ($r->{confLoc} =~ m/\,/);
		$match++;
		next unless ($locId = unique_name($r->{'confLoc'}));
		push(@updates, {
			'locId' => $locId,
			'id' => $r->{id}});
		print STDERR "Located $r->{confLoc}\n";
		if ($total++ >= $heur_limit) {
			$q_conference_todo->finish;
			last;
		}
	}
	print STDERR "Heuristic 2:\ntotal=$total match=$match nonunique=$nonunique unique=$unique notfound=$notfound\n";
	$dbh->do('BEGIN TRANSACTION');
	for my $urow (@updates) {
		$u->execute($urow->{'locId'}, $urow->{'id'}) or die;
	}
	$dbh->do('END TRANSACTION');
}

# Heuristic 3
# Conference location specified through a unique name of importance 1 or a country's capital
# Example: Paris
# Example: Budapest
sub
conf_name_3
{
	my $u = $dbh->prepare(q{
	  UPDATE Papers SET confLocId = ?, confLocHeur = 3 WHERE id = ?});
	$match = $total = $nonunique = $unique = $notfound = 0;
	$q_conference_todo->execute || die;
	my @updates;
	while (my $rp = $q_conference_todo->fetchrow_hashref) {
		$total++;
		next if ($r->{confLoc} =~ m/\,/);
		$match++;
		$q_important_name->execute($name = $rp->{'confLoc'});
		my $rw = fetchrow_count($q_important_name);
		if ($rw->{'n'} == 0) {
			logline("No important name: [$name]");
			$notfound++;
			next;
		} elsif ($rw->{'n'} > 1) {
			logline("No unique important name: [$name]");
			$nonunique++;
			next;
		}
		$unique++;
		push(@updates, {
			'locId' => $rw->{id},
			'id' => $rp->{id}});
		print STDERR "Located $rp->{confLoc}\n";
		if ($total++ >= $heur_limit) {
			$q_conference_todo->finish;
			last;
		}
	}
	print STDERR "Heuristic 3:\ntotal=$total match=$match nonunique=$nonunique unique=$unique notfound=$notfound\n";
	$dbh->do('BEGIN TRANSACTION');
	for my $urow (@updates) {
		$u->execute($urow->{'locId'}, $urow->{'id'}) or die;
	}
	$dbh->do('END TRANSACTION');
}

# Heuristic 6
# Conference location specified through a major city and country
# Example: Beijing, China
sub
conf_name_country_6
{
	my $u = $dbh->prepare(q{
	  UPDATE Papers SET confLocId = ?, confLocHeur = 6 WHERE id = ?});
	$total = $match_country = $nonunique = $unique = $notfound = $notfoundimp = $nonuniqueimp = $uniqueimp = 0;
	$q_conference_todo->execute || die;
	my @updates;
	while (my $rp = $q_conference_todo->fetchrow_hashref) {
		$total++;

		next unless ($rp->{confLoc} =~ m/^([^,]*)\,\s*(.*)$/);
		my $city = $1;
		my $country = $2;
		next if ($country =~ m/^[A-Z]{2}$/);	# State-like
		$match++;

		my $country_code = country_code($country);
		next unless defined($country_code);
		$match_country++;

		my $id = name_country($city, $country_code);
		next unless defined($id);
		push(@updates, {
			'locId' => $id,
			'id' => $rp->{id}});
		print STDERR "Located $rp->{confLoc}\n";
		if ($total++ >= $heur_limit) {
			$q_conference_todo->finish;
			last;
		}
	}
	print STDERR "Heuristic 6:\ntotal=$total match_country=$match_country nonunique=$nonunique unique=$unique notfound=$notfound notfoundimp=$notfoundimp nonuniqueimp=$nonuniqueimp  uniqueimp=$uniqueimp\n";
	$dbh->do('BEGIN TRANSACTION');
	for my $urow (@updates) {
		$u->execute($urow->{'locId'}, $urow->{'id'}) or die;
	}
	$dbh->do('END TRANSACTION');
}

# Remove postcode and space from a city's name
sub
clearup_city
{
	my ($city) = @_;

	# Remove postcodes
	$city =~ s/[A-Z]+\-\d+//;		# Europe: D-09107 Chemnitz
	$city =~ s/\d+\-\d+//;			# Japan, Poland, Brazil: Gunma 370-3347
	$city =~ s/[A-Z0-9]{3,4} [A-Z0-9]{3,4}//;# Britain and Canada: SW7 2BZ
	$city =~ s/\d+//g;			# All others

	# Remove leading and trailing garbage
	$city =~ s/^\s*\-?//;
	$city =~ s/\s*$//;
	return $city;
}

sub max
{
	my($a, $b) = @_;
	return $a > $b ? $a : $b;
}

# Return the administrative region of a Canadian address
# Removing it from the corresponding field
# Canada is a nightmare; the fields can be:
# Vancouver, Canada
# Vancouver, BC V6T 1Z3, Canada
# Vancouver, BC, V6S 2L9, Canada
# The area can be: Alta. Man. Ont. Que. Qu\x{e9}. Sask.
# AB AL BC NB NS QC ON
# See http://en.wikipedia.org/wiki/Canadian_subnational_postal_abbreviations
# We also have
# Vancouver, BC V6T 1Z4, United States and
# Cambridge, MA 02139, Canada
# but let's not get into that.
sub
canada_adm1
{
	my ($fieldref) = @_;

	my $nf = $#{$fieldref};
	for (my $i = $nf - 1; $i >= max($nf - 2, 0); $i--) {
		for my $sub (@canada_subs) {
			my $name = $sub->{'name'};
			if ($fieldref->[$i] =~ s/\b$name//) {
				return $sub->{'adm1'};
			}
		}
	}
	return undef;
}

if ('c' =~ m/$opt_t/) {
	print "Canadian provinces\n";
	my @fields = ('University of British Columbia', 'Vancouver', 'BC V6T 1Z4', 'Canada');
	print "Adm1=", canada_adm1(\@fields), " New fields=", join(',', @fields), "\n";
	my @fields = ('University of British Columbia', 'Vancouver', 'BC', 'V6T 1Z4', 'Canada');
	print "Adm1=", canada_adm1(\@fields), " New fields=", join(',', @fields), "\n";
	my @fields = ('University of British Columbia', 'Vancouver', 'BC', 'Canada');
	print "Adm1=", canada_adm1(\@fields), " New fields=", join(',', @fields), "\n";
	my @fields = ('Communications Research Centre', 'Ottawa', 'Ont.', 'Canada');
	print "Adm1=", canada_adm1(\@fields), " New fields=", join(',', @fields), "\n";
	# Unicode
	my @fields = split(/\,/, $dbh->selectrow_array(q{select corraddr from papers where eid='2-s2.0-2342510871'}));
	print "Original fields: ", join(',', @fields), "\n";
	print "Adm1=", canada_adm1(\@fields), " New fields=", join(',', @fields), "\n";
}

if ('u' =~ m/$opt_t/) {
	print "Unicode test\n";
	my ($canon) = $dbh->selectrow_array(q{select lname from World where extid=928702 and not nodiacritics;});
	print "Original $canon, No diacritics: ", strip_diacr($canon), "\n";
}

# Heuristic 4
# Author location specified through a city, possibly a state, and a country
# Example: Oslo, Norway
# Example: Livermore, CA 94550-9234, United States
sub
corr_name_country_4
{
	my $u = $dbh->prepare(q{
	  UPDATE Papers SET corrLocId = ?, corrLocHeur = 4 WHERE id = ?});
	$total = $match_country = $nonunique = $unique = $notfound = $notfoundimp = $nonuniqueimp = $uniqueimp = 0;
	$q_author_todo->execute || die;
	my @updates;
	while (my $r = $q_author_todo->fetchrow_hashref) {
		$total++;
		my $addr = $r->{corrAddr};
		print STDERR "Try:\t\t$addr\n";
		$addr =~ s/\;\s*email.*//;
		my @fields = split(/\,/, $addr);
		my $country = $fields[$#fields];
		my $country_code = country_code($country);
		next unless defined($country_code);
		$match_country++;

		my $state;
		$state = undef;
		if ($country_code eq 'CA') {
			$state = canada_adm1(\@fields);
		} elsif ($fields[$#fields - 1] =~ m/\b([A-Z][A-Z])\b/) {
			# US is very consistent
			$state = $1;
		}

		# Find a city name and clean it up
		my $city;
		for (my $i = $#fields - 1; $i >= 0; $i--) {
			if ($fields[$i] =~ m/[a-z]/) {
				$city = clearup_city($fields[$i]);
				last if ($city ne '');
			}
		}

		my $locId;
		if (defined($state)) {
			next unless ($locId = unique_name_state_country($city, $state, $country_code));
		} else {
			next unless ($locId = name_country($city, $country_code));
		}
		push(@updates, {
			'locId' => $locId,
			'id' => $r->{id}});
		print STDERR "Located:\t$r->{corrAddr}\n";
		if ($total++ >= $heur_limit) {
			$q_author_todo->finish;
			last;
		}
	}
	print STDERR "Heuristic 4:\ntotal=$total match_country=$match_country nonunique=$nonunique unique=$unique notfound=$notfound notfoundimp=$notfoundimp nonuniqueimp=$nonuniqueimp  uniqueimp=$uniqueimp\n";
	$dbh->do('BEGIN TRANSACTION');
	for my $urow (@updates) {
		$u->execute($urow->{'locId'}, $urow->{'id'}) or die;
	}
	$dbh->do('END TRANSACTION');
}
