#!/usr/bin/perl
#
# Calculate database overlap
#
use strict;

use File::Spec::Functions;
use IO::File;

use BibTeX::Parser;
use BibKeys;

use Data::Dumper;

use constant OVERLAP_DIR => "overlap";

use constant SCOPUS_SRCTYPE_P_2008 => 243595; # 24/03/2010

use constant RESULTS_FILE => "overlap_results.txt";

use constant Z_0025 => 1.96;

my @estimates;

opendir(DIR, OVERLAP_DIR);
my @bib_files = grep(/\.bib$/, readdir(DIR));
closedir(DIR);

my @isi_files = grep(/.*-isi\.bib/, @bib_files);
my @scopus_files = grep(/.*-scopus\.bib/, @bib_files);

open(my $results_fh, '>', RESULTS_FILE);

print $results_fh("isi:scopus:overlap:estimate\n");

for (my $i = 0; $i < @isi_files; $i++) {
  my $isi_file = @isi_files[$i];
  my $scopus_file = @scopus_files[$i];
  my $query = substr($isi_file, 0, index($isi_file, "-isi.bib", 0));
  if ($scopus_file !~ /$query/) {
    print STDOUT "mismatch $isi_file $scopus_file.\n";
    exit;
  }
  $isi_file = catfile(OVERLAP_DIR, $isi_file);
  $scopus_file = catfile(OVERLAP_DIR, $scopus_file);
  print STDOUT "matching $isi_file $scopus_file  ...\n";
  my ($overlap, $isi, $scopus) = match_files($isi_file, $scopus_file);
  if ($isi >= 50 && $scopus >= 50) {
    my $estimate = SCOPUS_SRCTYPE_P_2008 * $isi / $overlap;
    my $output = "$isi:$scopus:$overlap:$estimate\n";
    print STDOUT $output;
    print $results_fh $output;
    push(@estimates, $estimate);
  } else {
    print STDOUT "No estimate for $isi_file ($isi) $scopus_file ($scopus) \n";
  }
}

close($results_fh);

# The calculations below are derived by R. Montgomery's "Applied
# Statistics and Probability for Engineers", p.255, 3rd. edition,
# 2003.

my $std_dev = std_dev_ref_sum(\@estimates);
my $mean = mean_ref(\@estimates);

my $offset = Z_0025 * $std_dev / sqrt(@estimates);
my $lower_bound = $mean - $offset;
my $upper_bound = $mean + $offset;

print "mean=$mean with 95% CI [$lower_bound, $upper_bound] sample="
  . scalar(@estimates) . "\n";


# Utility functions

sub match_entry {
  my ($entry1, $entry2) = @_;
  (my $pages1 = $entry1->field('pages')) =~ s/--/-/;
  (my $pages2 = $entry2->field('pages')) =~ s/--/-/;
  if ($pages1 eq $pages2) {
    return 1;
  }
  return 0;
}



sub parse_file {
  my $fh = shift;

  my @entries;

  my $parser = BibTeX::Parser->new($fh);

  while (my $entry = $parser->next) {
    if (!$entry->parse_ok) {
      warn "Error parsing file: " . $entry->error;
    } else {
      push(@entries, $entry);
    }
  }

  return \@entries;
}

sub match_files {
  my ($isi_file, $scopus_file) = @_;

  my $fh = IO::File->new($isi_file);
  my @isi_entries = grep {$_->field('pages')} @{parse_file($fh)};
  my $num_isi_entries = scalar(@isi_entries);
  print "Parsed $num_isi_entries entries from $isi_file\n";
  $fh->close();

  $fh = IO::File->new($scopus_file, "<:via(BibKeys)");
  my @scopus_entries = grep {$_->field('pages') } @{parse_file($fh)};
  my $num_scopus_entries = scalar(@scopus_entries);
  $fh->close();
  print "Parsed $num_scopus_entries entries from $scopus_file\n";

  if ($num_isi_entries < 50 || $num_scopus_entries < 50) {
    return (0, $num_isi_entries, $num_scopus_entries);
  }

  my $matches = 0;
  for my $isi_entry (@isi_entries) {
    for my $scopus_entry (@scopus_entries) {
      my $match = match_entry($isi_entry, $scopus_entry);
      if ($match == 1) {
        #print "match: " . $isi_entry->key . " and " . $scopus_entry->key . "\n";
        $matches++;
      } 
    }
  }
  $num_isi_entries = scalar(@isi_entries);
  $num_scopus_entries = scalar(@scopus_entries);
  return ($matches, $num_isi_entries, $num_scopus_entries);
}

# Taken from: http://www.linuxjournal.com/article/6540
sub std_dev_ref_sum {
    my $ar = shift;
    my $elements = scalar @$ar;
    my $sum = 0;
    my $sumsq = 0;

    foreach (@$ar) {
        $sum += $_;
        $sumsq += ($_ **2);
    }
    return sqrt( $sumsq/$elements -
		 (($sum/$elements) ** 2));
}

# Taken from: http://www.linuxjournal.com/article/6540
sub mean_ref {
    my $result;
    my $ar = shift;
    foreach (@$ar) { $result += $_ }
    return $result / scalar(@$ar);
}

