#!/usr/bin/perl
#
# Convert tab-delimited files into TeX tables
#
# Diomidis Spinellis, February 2010
#
# $Id: tab2tex.pl,v 1.18 2013/04/19 11:53:05 dds Exp $
#

use Getopt::Std;
use File::Basename;

$opt_r = -1;
$opt_p = 2;
$opt_c = "";
$opt_s = 0;
$opt_w = "!ht";

getopts('c:H:h:l:p:r:t:') || die "usage: $0 [-c str] [-l str] [-p n] [-r n] [-t n] file ...
-c str  Table caption
-H str	Collon-separated list of level-1 column headings
-h str	Collon-separated list of level-2 column headings
-l str	Set table's label to str (default first table's base name)
-p n	Round columns titled % to so many digits (default 2)
-r n	Maximum number of rows
-t n	Convert text of column n (1-based) to title case
-w str  Where to put table (default !ht)
";

my @month_name = qw(
 NUL
 January February March
 April   May      June
 July    August   September
 October November December
);

$label = $opt_l if (defined($opt_l));

# Read tables and paste their lines together
while ($file = shift) {
	my @tlines;
	$label = fileparse($file, ('.txt')) unless (defined($label));
	open(IN, $file) || die "Unable to open $file: $!\n";
	@tlines = <IN>;

	if (!defined(@lines)) {
		@lines = @tlines;
		next;
	}

	# Append tlines to lines
	if ($#lines != $#tlines) {
		print STDERR "Table's $file lines are not the same number as those of previous ones'\n";
		exit 1;
	}
	for (my $i = 0; $i <= $#tlines; $i++) {
		$tlines[$i] =~ s/\n//;
		# An extra empty column is created to space clines
		$lines[$i] .= "\t\t" . $tlines[$i];
	}
}

# A double colon means a heading separation line
@HEADINGS2 = split(/:/, $opt_h);

# Create a HEADINGS2 top row, if needed
if ($opt_h =~ m /\\n/) {
	for ($i = 0; $i <= $#HEADINGS2; $i++) {
		push(@HEADINGS2_TOP, ($HEADINGS2[$i] =~ s/^(.*)\\n//) ? $1 : '');
	}
}

# Create table format
my $count_col_prev = 1;
my $count_col = 0;
for $col (split(/\t/, $lines[0])) {
	if ($HEADINGS2[$count_col] eq '') {
		$cline .= "\\cline{$count_col_prev-$count_col} ";
		$count_col_prev = $count_col + 2;
		$h1_columns[$h1_col++] = $h1_col_count;
		$h1_col_count = -1;
	}
	$h1_col_count++;
	if ($col =~ /^\d/ && $HEADINGS2[$count_col] ne 'Month') {
		$fmt .= 'r';
	} else {
		$fmt .= 'l';
	}
	$count_col++;
}
$h1_columns[$h1_col++] = $h1_col_count;
$cline .= "\\cline{$count_col_prev-$count_col}";

# Create heading1, if needed
if ($opt_H) {
	@HEADINGS1 = split(/:/, $opt_H);
	for ($i = 0; $i < $h1_col; $i++) {
		$h1 .= "\\multicolumn{$h1_columns[$i]}{l}{$HEADINGS1[$i]} ";
		$h1 .= '& & ' if ($i + 1 < $h1_col);
	}
	$h1 .= "\\\\\n";
} else {
	$cline = '';
}

print "\\begin{table}[$opt_w]
\\caption{}
\\begin{tabular}{$fmt}
$h1
$cline
";
row(@HEADINGS2_TOP) if ($#HEADINGS2_TOP > -1);
row(@HEADINGS2);
print "\\hline\n";

my $count_row = 1;

# Determine columns that can have their .0 removed
my @has_decimal_digits;
for my $line (@lines) {
	s/\n//;
	my $col_no = 0;
	for $_ (split(/\t/, $line)) {
		$has_decimal_digits[$col_no]++ if (/\.[1-9]+$/);
		$col_no++;
	}
        last if $count_row++ == $opt_r;
}


$count_row = 1;
for $_ (@lines) {
	s/\n//;
	@cols = split(/\t/, $_);
	row(@cols);
        last if $count_row++ == $opt_r;
}
print "
\\hline
\\end{tabular}
\\begin{flushleft}$opt_c
\\end{flushleft}
\\label{somtab:$label}
\\end{table}
";

sub row
{
	my(@cols) = @_;
	# Convert to title case
	$cols[$opt_t - 1] =~ s/\b(\w)(\w+)\b/$1 . lc($2)/eg if ($opt_t);
	for (my $i = 0; $i <= $#cols; $i++) {
		# Handle rounding
		if ($has_decimal_digits[$i]) {
			if ($HEADINGS2[$i] =~ m/\%/) {
				# Round percentages to two decimals
				$cols[$i] = sprintf("%.${opt_p}f", $cols[$i]);
			} else {
				# Round all else to integers
				$cols[$i] = sprintf("%.0f", $cols[$i]);
			}
		} else {
			# Remove trailing .0
			$cols[$i] =~ s/\.0$// unless ($has_decimal_digits[$i]);
		}
		# Add thousands separator
		$cols[$i] =~ s/(\d{1,3}?)(?=(\d{3})+$)/$1,/g
			unless ($HEADINGS2[$i] eq 'Year');
		$cols[$i] = $month_name[$cols[$i]]
			if ($HEADINGS2[$i] eq 'Month' && $cols[$i] =~ m/^\d+$/);
	}
	  print join(' & ', @cols), " \\\\\n";
}

# Local Variables:
# indent-tabs-mode:t
# cperl-indent-level:8
# End:
